/***********************************************************************
LeapNavigationTool - Class encapsulating the navigation behaviour of a
mouse in the OpenInventor SoXtExaminerViewer.
Copyright (c) 2004-2011 Oliver Kreylos

This file is part of the Virtual Reality User Interface Library (Vrui).

The Virtual Reality User Interface Library is free software; you can
redistribute it and/or modify it under the terms of the GNU General
Public License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

The Virtual Reality User Interface Library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU General Public License for more details.
96352.1
You should have received a copy of the GNU General Public License along
with the Virtual Reality User Interface Library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA
***********************************************************************/

#include "LeapNavigationTool.h"

#include <Misc/StandardValueCoders.h>
#include <Misc/ConfigurationFile.h>
#include <Math/Math.h>
#include <Geometry/GeometryValueCoders.h>
#include <GL/gl.h>
#include <GL/GLColorTemplates.h>
#include <GL/GLGeometryWrappers.h>
#include <GL/GLTransformationWrappers.h>
#include <Vrui/Vrui.h>
#include <Vrui/InputGraphManager.h>
#include <Vrui/InputDeviceManager.h>
#include <Vrui/Viewer.h>
#include <Vrui/VRWindow.h>
#include <Vrui/ToolManager.h>
#include <Vrui/VRScreen.h>

#include <linux/input.h>

const std::string fingerNames[] = {"Thumb", "Index", "Middle", "Ring", "Pinky"};
const std::string boneNames[] = {"Metacarpal", "Proximal", "Middle", "Distal"};
const std::string stateNames[] = {"STATE_INVALID", "STATE_START", "STATE_UPDATE", "STATE_END"};

namespace Vrui {

  /*******************************************
Methods of class LeapNavigationToolFactory:
  *******************************************/

  LeapNavigationToolFactory::LeapNavigationToolFactory(ToolManager& toolManager)
    :ToolFactory("LeapNavigationTool",toolManager),
     rotatePlaneOffset(getInchFactor()*Scalar(3)),
     rotateFactor(getInchFactor()*Scalar(3)),
     invertDolly(false),
     screenDollyingDirection(0,-1,0),
     screenScalingDirection(0,-1,0),
     dollyFactor(Scalar(1)),
     scaleFactor(getInchFactor()*Scalar(3)),
     wheelDollyFactor(getInchFactor()*Scalar(-12)),
     wheelScaleFactor(Scalar(0.5)),
     spinThreshold(getUiSize()*Scalar(1)),
     showScreenCenter(true),
     interactWithWidgets(true)
  {
    /* Initialize tool layout: */
    layout.setNumButtons(1);
    layout.setNumValuators(0);
	
    /* Insert class into class hierarchy: */
    ToolFactory* navigationToolFactory=toolManager.loadClass("NavigationTool");
    navigationToolFactory->addChildClass(this);
    addParentClass(navigationToolFactory);
	
    /* Load class settings: */
    Misc::ConfigurationFileSection cfs=toolManager.getToolClassSection(getClassName());
    rotatePlaneOffset=cfs.retrieveValue<Scalar>("./rotatePlaneOffset",rotatePlaneOffset);
    rotateFactor=cfs.retrieveValue<Scalar>("./rotateFactor",rotateFactor);
    invertDolly=cfs.retrieveValue<bool>("./invertDolly",invertDolly);
    screenDollyingDirection=cfs.retrieveValue<Vector>("./screenDollyingDirection",screenDollyingDirection);
    screenScalingDirection=cfs.retrieveValue<Vector>("./screenScalingDirection",screenScalingDirection);
    dollyFactor=cfs.retrieveValue<Scalar>("./dollyFactor",dollyFactor);
    scaleFactor=cfs.retrieveValue<Scalar>("./scaleFactor",scaleFactor);
    wheelDollyFactor=cfs.retrieveValue<Scalar>("./wheelDollyFactor",wheelDollyFactor);
    wheelScaleFactor=cfs.retrieveValue<Scalar>("./wheelScaleFactor",wheelScaleFactor);
    spinThreshold=cfs.retrieveValue<Scalar>("./spinThreshold",spinThreshold);
    showScreenCenter=cfs.retrieveValue<bool>("./showScreenCenter",showScreenCenter);
    interactWithWidgets=cfs.retrieveValue<bool>("./interactWithWidgets",interactWithWidgets);
	
    /* Set tool class' factory pointer: */
    LeapNavigationTool::factory=this;
  }

  LeapNavigationToolFactory::~LeapNavigationToolFactory(void)
  {
    /* Reset tool class' factory pointer: */
    LeapNavigationTool::factory=0;
  }

  const char* LeapNavigationToolFactory::getName(void) const
  {
    return "Leap Navigation";
  }

  Tool* LeapNavigationToolFactory::createTool(const ToolInputAssignment& inputAssignment) const
  {
    return new LeapNavigationTool(this,inputAssignment);
  }

  void LeapNavigationToolFactory::destroyTool(Tool* tool) const
  {
    delete tool;
  }

  extern "C" void resolveLeapNavigationToolDependencies(Plugins::FactoryManager<ToolFactory>& manager)
  {
    /* Load base classes: */
    manager.loadClass("NavigationTool");
  }

  extern "C" ToolFactory* createLeapNavigationToolFactory(Plugins::FactoryManager<ToolFactory>& manager)
  {
    /* Get pointer to tool manager: */
    ToolManager* toolManager=static_cast<ToolManager*>(&manager);
	
    /* Create factory object and insert it into class hierarchy: */
    LeapNavigationToolFactory* mouseNavigationToolFactory=new LeapNavigationToolFactory(*toolManager);
	
    /* Return factory object: */
    return mouseNavigationToolFactory;
  }

  extern "C" void destroyLeapNavigationToolFactory(ToolFactory* factory)
  {
    delete factory;
  }

  /********************************************
Static elements of class LeapNavigationTool:
  ********************************************/

  LeapNavigationToolFactory* LeapNavigationTool::factory=0;

  /************************************
Methods of class LeapNavigationTool:
  ************************************/

  LeapNavigationTool::LeapNavigationTool(const ToolFactory* factory,const ToolInputAssignment& inputAssignment)
    :NavigationTool(factory,inputAssignment),
     GUIInteractor(false,Scalar(0),getButtonDevice(0)),
     mouseAdapter(0),
     currentPos(Point::origin),currentValue(0),
     dolly(LeapNavigationTool::factory->invertDolly),navigationMode(IDLE)
  {
    //pthread_create (&background, 0, backgroundthread, this);
    startLeap ();
  }

  const ToolFactory* LeapNavigationTool::getFactory(void) const
  {
    return factory;
  }

  Point LeapNavigationTool::calcScreenCenter(void) const
  {
    /* Get the transformation of the screen currently containing the input device: */
    Scalar viewport[4];
    ONTransform screenT=getScreenTransform(viewport);
	
    /* Calculate the screen's center: */
    Point center;
    center[0]=Math::mid(viewport[0],viewport[1]);
    center[1]=Math::mid(viewport[2],viewport[3]);
    center[2]=Scalar(0);
	
    /* Transform the center position to physical coordinates: */
    return screenT.transform(center);
  }

  ONTransform LeapNavigationTool::getScreenTransform(Scalar viewport[4]) const
  {
    /* Check if the mouse adapter is valid: */
    VRScreen* screen=0;

    /* Use the main screen: */
    //screen=getMainScreen();
    screen = getScreen (getNumScreens () - 1);
    screen->getViewport(viewport);

    /* Return the screen's transformation: */
    return screen->getScreenTransformation();
  }

  void LeapNavigationTool::frame(void)
  {
    if (listener.fingercount == 0) {
      navigationMode=IDLE;
      return;
    }

    static Leap::Vector oldfinger1, oldfinger2, olddirection,
      oldthumb, oldindex;
    static NavTrackerState inverse;

    if (listener.fingercount == 1) {
      if (navigationMode != PANNING) {
	oldfinger1 = listener.finger1;
	inverse = NavTrackerState::identity;
	olddirection = listener.direction;
	navigationMode = PANNING;
      }
    } else if (listener.fingercount == 2) {
      if (navigationMode != SCALING) {
	oldfinger1 = listener.finger1;
	oldfinger2 = listener.finger2;
	inverse = NavTrackerState::identity;
	navigationMode = SCALING;
      }
    } else if (listener.pinch) {
      if (navigationMode != ROTATING) {
	oldthumb = listener.thumb;
	oldindex = listener.index;
	navigationMode = ROTATING;
      }
    } else {
      navigationMode=IDLE;
    }

    /* Act depending on this tool's current state: */
    switch(navigationMode)
      {
      case SCALING:
	{
	  float scale =  (listener.finger1 - listener.finger2).magnitude () /
	    (oldfinger1 - oldfinger2).magnitude ();

	  concatenateNavigationTransformationLeft (inverse);

	  Point offset = calcScreenCenter ();
	  NavTrackerState offsetT, inverseOffsetT, t;

	  offsetT = NavTrackerState::translateFromOriginTo(offset);
	  inverseOffsetT = NavTrackerState::translateToOriginFrom(offset);

	  t = offsetT;
	  t *= NavTrackerState::scale (scale);
	  t *= inverseOffsetT;

	  inverse = inverseOffsetT;
	  inverse *= NavTrackerState::scale (1.0f / scale);
	  inverse *= offsetT;

	  concatenateNavigationTransformationLeft (t);

	  break;
	}
		
      case PANNING:
	{
	  /* Update the navigation transformation: */
	  concatenateNavigationTransformationLeft(inverse);

	  Leap::Vector tmp = 0.03 * (listener.finger1 - oldfinger1);

	  NavTrackerState t =
	    NavTrackerState::translate(Vector (tmp[0], tmp[2], tmp[1]));

	  inverse = 
	    NavTrackerState::translate(Vector (-tmp[0], -tmp[2], -tmp[1]));

	  concatenateNavigationTransformationLeft(t);
	  break;
	}
		
      case ROTATING:
	{
	  /* Update the navigation transformation: */
	  float angle =
	    -2.0 * (listener.thumb - listener.index).angleTo (oldthumb -
							      oldindex);
	  Leap::Vector cross =
	    (listener.thumb - listener.index).cross (oldthumb -
						     oldindex);

	  concatenateNavigationTransformationLeft(inverse);

	  Point offset = calcScreenCenter ();
	  NavTrackerState offsetT, inverseOffsetT, t;

	  offsetT = NavTrackerState::translateFromOriginTo(offset);
	  inverseOffsetT = NavTrackerState::translateToOriginFrom(offset);

	  t = offsetT;
	  t *= NavTrackerState::rotate(NavTrackerState::Rotation::rotateAxis(Vector (cross[0], cross[2], cross[1]), angle));
	  t *= inverseOffsetT;

	  inverse = inverseOffsetT;
	  inverse *= NavTrackerState::rotate(NavTrackerState::Rotation::rotateAxis(Vector (cross[0], cross[2], cross[1]), -angle));
	  inverse *= offsetT;

	  concatenateNavigationTransformationLeft(t);
	  break;
	}
		
      default:
	;
      }
  }

  void LeapNavigationTool::startLeap ()
  {
    // Have the sample listener receive events from the controller
    controller.addListener(listener);

    controller.setPolicy(Leap::Controller::POLICY_BACKGROUND_FRAMES);
  }

  void LeapNavigationTool::stopLeap ()
  {
    // Remove the sample listener when done
    controller.removeListener(listener);
 }

  void LeapNavigationTool::display(GLContextData& contextData) const
  {
  }

  void LeapNavigationTool::createLeapTool ()
  {
    Vrui::LeapNavigationToolFactory *toolfactory;
    Vrui::ToolInputLayout layout;
    Vrui::ToolInputAssignment assignment (layout);
    Vrui::ToolManager *toolmanager = Vrui::getToolManager ();

    toolfactory =
      new Vrui::LeapNavigationToolFactory (*toolmanager);

    const Leap::DeviceList devices = LeapNavigationTool::controller.devices();

    if (devices.count ()) {
      toolmanager->createTool (toolfactory, assignment);
    }
    else
      toolmanager->addClass (toolfactory);
  }

  void LeapNavigationTool::SampleListener::onInit(const Leap::Controller& controller) {
    std::cout << "Initialized" << std::endl;
  }

  void LeapNavigationTool::SampleListener::onConnect(const Leap::Controller& controller) {
    std::cout << "Connected" << std::endl;
#if 0
    controller.enableGesture(Leap::Gesture::TYPE_CIRCLE);
    controller.enableGesture(Leap::Gesture::TYPE_KEY_TAP);
    controller.enableGesture(Leap::Gesture::TYPE_SCREEN_TAP);
    controller.enableGesture(Leap::Gesture::TYPE_SWIPE);
#endif
  }

  void LeapNavigationTool::SampleListener::onDisconnect(const Leap::Controller& controller) {
    // Note: not dispatched when running in a debugger.
    std::cout << "Disconnected" << std::endl;
  }

  void LeapNavigationTool::SampleListener::onExit(const Leap::Controller& controller) {
    std::cout << "Exited" << std::endl;
  }

  void LeapNavigationTool::SampleListener::onFrame(const Leap::Controller& controller) {
    // Get the most recent frame and report some basic information
    frame = controller.frame();

    fingers = frame.fingers().extended();
    fingercount = fingers.count();

    Leap::HandList hands = frame.hands();
    if (fingercount == 2) {
      Leap::FingerList::const_iterator fl = fingers.begin();
      finger1 = (*fl).tipPosition ();
      ++fl;
      finger2 = (*fl).tipPosition ();
    } else if (fingercount == 1) {
      Leap::FingerList::const_iterator fl = fingers.begin();
      finger1 = (*fl).tipPosition ();
      for (Leap::HandList::const_iterator hl = hands.begin();
	   hl != hands.end(); ++hl) {
	const Leap::Hand hand = *hl;
	for (Leap::FingerList::const_iterator fl =
	       hand.fingers ().extended ().begin ();
	     fl != hand.fingers ().extended ().end (); ++fl) {
	  if (hand.fingers ().extended ().count ()) {
	    direction = hand.direction ();
	    break;
	  }
	}
      }
    }

    for (Leap::HandList::const_iterator hl = hands.begin();
	 hl != hands.end(); ++hl) {
      const Leap::Hand hand = *hl;
      int count = 0;
      if (hand.fingers ().extended ().count ()) {
	const Leap::FingerList fingers = hand.fingers();
	for (Leap::FingerList::const_iterator fl =
	       fingers.begin ();
	     fl != fingers.end (); ++fl) {
	  const Leap::Finger finger = *fl;
	  if (finger.type () == Leap::Finger::TYPE_THUMB) {
	    thumb = finger.tipPosition ();
	    count++;
	  }
	  if (finger.type () == Leap::Finger::TYPE_INDEX) {
	    index = finger.tipPosition ();
	    count++;
	  }
	}
	if (count == 2)
	  pinch = 1;
	else
	  pinch = 0;
      }
    }

    Vrui::requestUpdate ();
    return;

    std::cout << "Frame id: " << frame.id()
	      << ", timestamp: " << frame.timestamp()
	      << ", hands: " << frame.hands().count()
	      << ", extended fingers: " << frame.fingers().extended().count()
	      << ", tools: " << frame.tools().count()
	      << ", gestures: " << frame.gestures().count() << std::endl;

    //Leap::HandList hands = frame.hands();
    for (Leap::HandList::const_iterator hl = hands.begin(); hl != hands.end(); ++hl) {
      // Get the first hand
      const Leap::Hand hand = *hl;
      std::string handType = hand.isLeft() ? "Left hand" : "Right hand";
      std::cout << std::string(2, ' ') << handType << ", id: " << hand.id()
		<< ", palm position: " << hand.palmPosition() << std::endl;
      // Get the hand's normal vector and direction
      const Leap::Vector normal = hand.palmNormal();
      const Leap::Vector direction = hand.direction();

      // Calculate the hand's pitch, roll, and yaw angles
      std::cout << std::string(2, ' ') <<  "pitch: " << direction.pitch() * Leap::RAD_TO_DEG << " degrees, "
		<< "roll: " << normal.roll() * Leap::RAD_TO_DEG << " degrees, "
		<< "yaw: " << direction.yaw() * Leap::RAD_TO_DEG << " degrees" << std::endl;

      // Get the Arm bone
      Leap::Arm arm = hand.arm();
      std::cout << std::string(2, ' ') <<  "Arm direction: " << arm.direction()
		<< " wrist position: " << arm.wristPosition()
		<< " elbow position: " << arm.elbowPosition() << std::endl;

      // Get fingers
      const Leap::FingerList fingers = hand.fingers();
      for (Leap::FingerList::const_iterator fl = fingers.begin(); fl != fingers.end(); ++fl) {
	const Leap::Finger finger = *fl;
	std::cout << std::string(4, ' ') <<  fingerNames[finger.type()]
		  << " finger, id: " << finger.id()
		  << ", length: " << finger.length()
		  << "mm, width: " << finger.width() << std::endl;

	// Get finger bones
	for (int b = 0; b < 4; ++b) {
	  Leap::Bone::Type boneType = static_cast<Leap::Bone::Type>(b);
	  Leap::Bone bone = finger.bone(boneType);
	  std::cout << std::string(6, ' ') <<  boneNames[boneType]
		    << " bone, start: " << bone.prevJoint()
		    << ", end: " << bone.nextJoint()
		    << ", direction: " << bone.direction() << std::endl;
	}
      }
    }

    // Get tools
    const Leap::ToolList tools = frame.tools();
    for (Leap::ToolList::const_iterator tl = tools.begin(); tl != tools.end(); ++tl) {
      const Leap::Tool tool = *tl;
      std::cout << std::string(2, ' ') <<  "Tool, id: " << tool.id()
		<< ", position: " << tool.tipPosition()
		<< ", direction: " << tool.direction() << std::endl;
    }

    // Get gestures
    const Leap::GestureList gestures = frame.gestures();
    for (int g = 0; g < gestures.count(); ++g) {
      Leap::Gesture gesture = gestures[g];

      switch (gesture.type()) {
      case Leap::Gesture::TYPE_CIRCLE:
	{
	  Leap::CircleGesture circle = gesture;
	  std::string clockwiseness;

	  if (circle.pointable().direction().angleTo(circle.normal()) <= Leap::PI/2) {
	    clockwiseness = "clockwise";
	  } else {
	    clockwiseness = "counterclockwise";
	  }

	  // Calculate angle swept since last frame
	  float sweptAngle = 0;
	  if (circle.state() != Leap::Gesture::STATE_START) {
	    Leap::CircleGesture previousUpdate = Leap::CircleGesture(controller.frame(1).gesture(circle.id()));
	    sweptAngle = (circle.progress() - previousUpdate.progress()) * 2 * Leap::PI;
	  }
	  std::cout << std::string(2, ' ')
		    << "Circle id: " << gesture.id()
		    << ", state: " << stateNames[gesture.state()]
		    << ", progress: " << circle.progress()
		    << ", radius: " << circle.radius()
		    << ", angle " << sweptAngle * Leap::RAD_TO_DEG
		    <<  ", " << clockwiseness << std::endl;
	  break;
	}
      case Leap::Gesture::TYPE_SWIPE:
	{
	  Leap::SwipeGesture swipe = gesture;
	  std::cout << std::string(2, ' ')
		    << "Swipe id: " << gesture.id()
		    << ", state: " << stateNames[gesture.state()]
		    << ", direction: " << swipe.direction()
		    << ", speed: " << swipe.speed() << std::endl;
	  break;
	}
      case Leap::Gesture::TYPE_KEY_TAP:
	{
	  Leap::KeyTapGesture tap = gesture;
	  std::cout << std::string(2, ' ')
		    << "Key Tap id: " << gesture.id()
		    << ", state: " << stateNames[gesture.state()]
		    << ", position: " << tap.position()
		    << ", direction: " << tap.direction()<< std::endl;
	  break;
	}
      case Leap::Gesture::TYPE_SCREEN_TAP:
	{
	  Leap::ScreenTapGesture screentap = gesture;
	  std::cout << std::string(2, ' ')
		    << "Screen Tap id: " << gesture.id()
		    << ", state: " << stateNames[gesture.state()]
		    << ", position: " << screentap.position()
		    << ", direction: " << screentap.direction()<< std::endl;
	  break;
	}
      default:
        std::cout << std::string(2, ' ')  << "Unknown gesture type." << std::endl;
        break;
      }
    }

    if (!frame.hands().isEmpty() || !gestures.isEmpty()) {
      std::cout << std::endl;
    }

  }

  void LeapNavigationTool::SampleListener::onFocusGained(const Leap::Controller& controller) {
    std::cout << "Focus Gained" << std::endl;
  }

  void LeapNavigationTool::SampleListener::onFocusLost(const Leap::Controller& controller) {
    std::cout << "Focus Lost" << std::endl;
  }

  void LeapNavigationTool::SampleListener::onDeviceChange(const Leap::Controller& controller) {
    std::cout << "Device Changed" << std::endl;
    const Leap::DeviceList devices = controller.devices();

    for (int i = 0; i < devices.count(); ++i) {
      std::cout << "id: " << devices[i].toString() << std::endl;
      std::cout << "  isStreaming: " << (devices[i].isStreaming() ? "true" : "false") << std::endl;
    }
  }

  void LeapNavigationTool::SampleListener::onServiceConnect(const Leap::Controller& controller) {
    std::cout << "Service Connected" << std::endl;
  }

  void LeapNavigationTool::SampleListener::onServiceDisconnect(const Leap::Controller& controller) {
    std::cout << "Service Disconnected" << std::endl;
  }
}

Leap::Controller Vrui::LeapNavigationTool::controller;
