/***********************************************************************
LightSaberTool - Class for tools using light sabers to point out features in a
3D display.
Copyright (c) 2007-2011 Oliver Kreylos

This file is part of the Virtual Reality User Interface Library (Vrui).

The Virtual Reality User Interface Library is free software; you can
redistribute it and/or modify it under the terms of the GNU General
Public License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

The Virtual Reality User Interface Library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with the Virtual Reality User Interface Library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA
***********************************************************************/

#ifndef VRUI_LIGHTSABERTOOL_INCLUDED
#define VRUI_LIGHTSABERTOOL_INCLUDED

#include <string>
#include <Geometry/Point.h>
#include <Geometry/Vector.h>
#include <GL/gl.h>
#include <GL/GLObject.h>
#include <Images/RGBImage.h>
#include <Vrui/Geometry.h>
#include <Vrui/TransparentObject.h>
#include <Vrui/PointingTool.h>
#include <Sound/SoundPlayer.h>

#include <AL/al.h>
#include <AL/ALGeometryWrappers.h>
#include <AL/ALObject.h>
#include <AL/ALContextData.h>

namespace Vrui {

class LightSaberTool;

class LightSaberToolFactory:public ToolFactory
	{
	friend class LightSaberTool;
	
	/* Elements: */
	private:
	Scalar lightsaberLength; // Length of light saber billboard
	Scalar lightsaberWidth; // Width of light saber billboard
	Scalar baseOffset; // Amount by how much the light saber billboard is shifted towards the hilt
	std::string lightsaberImageFileName; // Name of image file containing light saber texture
	
	/* Constructors and destructors: */
	public:
	LightSaberToolFactory(ToolManager& toolManager);
	virtual ~LightSaberToolFactory(void);
	
	/* Methods from ToolFactory: */
	virtual const char* getName(void) const;
	virtual const char* getButtonFunction(int buttonSlotIndex) const;
	virtual Tool* createTool(const ToolInputAssignment& inputAssignment) const;
	virtual void destroyTool(Tool* tool) const;
	};

class LightSaberTool:public PointingTool,public GLObject,public TransparentObject
	{
	friend class LightSaberToolFactory;
	
	/* Embedded classes: */
	private:
	struct DataItem:public GLObject::DataItem
		{
		/* Elements: */
		public:
		GLuint textureObjectId; // ID of the light saber texture object
		
		/* Constructors and destructors: */
		DataItem(void)
			{
			glGenTextures(1,&textureObjectId);
			}
		virtual ~DataItem(void)
			{
			glDeleteTextures(1,&textureObjectId);
			}
		};
	
	/* Elements: */
	static LightSaberToolFactory* factory; // Pointer to the factory object for this class
	Images::RGBImage lightsaberImage; // The light saber texture image
	
	/* Transient state: */
	bool active; // Flag if the light saber is active
	double activationTime; // Time at which the light saber was activated
	Point origin; // Origin point of the light saber blade
	Vector axis; // Current light saber blade axis vector
	Scalar length; // Current light saber blade length

	static LightSaberTool *lightsaber; // reference to the last one we created

	// various sound effects
#if ALSUPPORT_CONFIG_HAVE_OPENAL
	ALuint turnon, turnoff, hit;
	ALuint turnonbuffer, turnoffbuffer, hitbuffer;
#endif
	
	/* Constructors and destructors: */
	public:
	LightSaberTool(const ToolFactory* factory,const ToolInputAssignment& inputAssignment);

	~LightSaberTool ();

	Point& getCurrentOrigin () { return origin; }
	Vector& getCurrentAxis () { return axis; }
	Scalar& getBladeLength () { return length; }
	bool isActive () { return active; }

	static LightSaberTool* getTool () { return lightsaber; }
	
	/* Methods from Tool: */
	virtual const ToolFactory* getFactory(void) const;
	virtual void buttonCallback(int buttonSlotIndex,InputDevice::ButtonCallbackData* cbData);
	virtual void frame(void);
	
	/* Methods from GLObject: */
	virtual void initContext(GLContextData& contextData) const;
	
	/* Methods from TransparentObject: */
	virtual void glRenderActionTransparent(GLContextData& contextData) const;

	static ALenum GetFormatFromInfo(short channels, short bitsPerSample) {
	  if (channels == 1) {
	    if (bitsPerSample == 16)
	      return AL_FORMAT_MONO16;
	    else
	      return AL_FORMAT_MONO8;
	  }
	  if (bitsPerSample == 16)
	    return AL_FORMAT_STEREO16;
	  return  AL_FORMAT_STEREO8;
	}

	static void swap (char *c1, char *c2) {
	  char c;

	  c = *c1;
	  *c1 = *c2;
	  *c2 = c;
	}

	static int file_read_int32_le (char *buffer, FILE* file) {
	  int result;

	  fread (buffer, 1, 4, file);
	  //swap (buffer, &(buffer[3]));
	  //swap (&(buffer[1]), &(buffer[2]));

	  result = *((int *)buffer);
	  return result;
	}

	static unsigned int file_read_uint32_le (char *buffer, FILE* file) {
	  unsigned int result;

	  fread (buffer, 1, 4, file);
	  //swap (buffer, &(buffer[3]));
	  //swap (&(buffer[1]), &(buffer[2]));

	  result = *((unsigned int *)buffer);
	  return result;
	}

	static short file_read_int16_le (char *buffer, FILE* file) {
	  short result;

	  fread (buffer, 1, 2, file);
	  //swap (buffer, &(buffer[1]));

	  result = *((short *)buffer);
	  return result;
	}

	static void loadALBuffer(const char *path, ALuint &source, ALuint &buffer)
	  {
	    int result;

	    // Load wav data into a buffers.
	    FILE* file = fopen(path, "rb");
	    if (!file) {
	      printf ("could not open file %s\n", path);
	      return;
	    }
	    char xbuffer[5];
	    xbuffer[4] = '\0';
	    if (fread(xbuffer, sizeof(char), 4, file) != 4 || strcmp(xbuffer, "RIFF") != 0)
	      printf ("%s is not a WAV file\n", path);

	    fread (xbuffer, 1, 4, file);

	    if (fread(xbuffer, sizeof(char), 4, file) != 4 || strcmp(xbuffer, "WAVE") != 0)
	      printf ("%s is not a WAV file\n", path);

	    if (fread(xbuffer, sizeof(char), 4, file) != 4 || strcmp(xbuffer, "fmt ") != 0)
	      printf ("%s is not a WAV file\n", path);

	    file_read_int32_le(xbuffer, file);
	    short audioFormat = file_read_int16_le(xbuffer, file);
	    short channels = file_read_int16_le(xbuffer, file);
	    int sampleRate = file_read_int32_le(xbuffer, file);
	    int byteRate = file_read_int32_le(xbuffer, file);
	    file_read_int16_le(xbuffer, file);
	    short bitsPerSample = file_read_int16_le(xbuffer, file);

#if 0
	    if (audioFormat != 16) {
	      short extraParams = file_read_int16_le(xbuffer, file);
	      for (int i=0; i<extraParams; i++)
		fread (xbuffer, 1, 1, file);
	    }
#endif

	    if (fread(xbuffer, sizeof(char), 4, file) != 4 || strcmp(xbuffer, "data") != 0)
	      printf ("Invalid WAV file\n");

	    int dataChunkSize = file_read_uint32_le(xbuffer, file);
	    void *bufferData = malloc (dataChunkSize);
	    fread (bufferData, 1, dataChunkSize, file);

	    float duration = float(dataChunkSize) / byteRate;
	    alBufferData(buffer, GetFormatFromInfo(channels, bitsPerSample), bufferData, dataChunkSize, sampleRate);
	    free(bufferData);
	    fclose(file);
	  }

	const bool isPlaying (ALuint &source) const
	{
	  ALenum state;
    
	  alGetSourcei(source, AL_SOURCE_STATE, &state);
    
	  return (state == AL_PLAYING);
	}

	};

}

#endif
